(* Time-stamp: <modified the 07/10/2021 (at 16:54) by Erwan Jahier> *)


type t = {
  mutable topo: string;
  mutable length: int;
  mutable cores_nb: int;
  mutable verbose: int;
  mutable daemon: DaemonType.t;
  mutable rif: bool;
  mutable no_data_file: bool;
  mutable quiet: bool;
  mutable ifi: bool;
  mutable gen_lutin: bool;
  mutable gen_oracle: bool;
  mutable dummy_input: bool;
  mutable output_algos: bool;
  mutable gen_register: bool;
  mutable init_search_max_trials: int option;
  
  mutable _args : (string * Arg.spec * string) list;
  mutable _user_man  : (string * string list) list; 
  mutable _hidden_man: (string * string list) list; 

  mutable _others : string list;
  mutable _margin : int;
}

let usage_msg tool = ("usage: " ^ tool ^ " [<option>]* <topology>.dot 
use -h to see the available options.
" )
let print_usage tool = Printf.printf "%s\n" (usage_msg tool); flush stdout


let (make_args : unit -> t) = 
  fun () -> 
    {
      topo = "";
      length = 10000;
      cores_nb = 1;
      verbose = 0;
      daemon = DaemonType.Distributed;
      rif = false;
      no_data_file = false;
      quiet = false;
      ifi = false;
      gen_lutin = false;
      gen_oracle = false;
      dummy_input = false;
      output_algos = false;
      gen_register = false;
      init_search_max_trials = None;
      _args = [];        
      _user_man  = [];   
      _hidden_man  = []; 
      _others = [];
      _margin =12;
    }


let pspec args os  (c, ml) = (
  let (m1, oth) = match ml with
	 |	h::t -> (h,t)
	 |	_ -> ("",[])
  in
  let t2 = String.make args._margin ' ' in
  let cl = String.length c in
  let t1 = if (cl < args._margin ) then
	 String.make (args._margin - cl) ' '
  else
	 "\n"^t2
  in
	 Printf.fprintf os "%s%s%s" c t1 m1;
	 List.iter (function x -> Printf.fprintf os "\n%s%s" t2 x) oth ;
	 Printf.fprintf os "\n" ;
)

let options args oc = (
	let l = List.rev args._user_man in
	List.iter (pspec args oc) l
)
let more_options args oc = (
	let l = List.rev (args._hidden_man) in
	List.iter (pspec args oc) l
)
let (mkopt : t -> string list -> ?hide:bool -> ?arg:string -> Arg.spec ->
     string list -> unit) =
  fun opt ol ?(hide=false) ?(arg="") se ml ->
    let treto o = opt._args <- (o, se, "")::opt._args in
	   List.iter treto ol ;
	   let col1 = (String.concat ", " ol)^arg in
	     if hide 
        then opt._hidden_man <- (col1, ml)::opt._hidden_man
	     else opt._user_man   <- (col1, ml)::opt._user_man

let myexit i = exit i
(*******************************************************************************)

(*** User Options Tab **)
let (mkoptab : string array -> t -> unit) = 
  fun argv args ->  
    (
    mkopt args  ["--synchronous-daemon";"-sd"]
      (Arg.Unit(fun () -> args.daemon <- DaemonType.Synchronous))
      ["Use a Synchronous daemon"];

    mkopt args  ["--central-daemon";"-cd"]
      (Arg.Unit(fun () -> args.daemon <- DaemonType.Central))
      ["Use a Central daemon (selects exactly one action)"];
    
    mkopt args  ["--locally-central-daemon";"-lcd"]
      (Arg.Unit(fun () -> args.daemon <- DaemonType.LocallyCentral))
      ["Use a Locally Central daemon";
       "(i.e., never activates two neighbors actions in the same step)"];
    mkopt args  ["--distributed-daemon";"-dd"]
      (Arg.Unit(fun () -> args.daemon <- DaemonType.Distributed))
      ["Use a Distributed daemon (which select at least one action).";
       "This is the default daemon."];
    
    mkopt args  ["--custom-daemon";"-custd"]
      (Arg.Unit(fun () -> args.daemon <- DaemonType.Custom;args.rif <- true))
      ["Use a Custom daemon (forces --rif)"];

    mkopt args  ["--greedy-central-daemon";"-gcd"]
      (Arg.Unit(fun () -> args.daemon <- DaemonType.GreedyCentral))
      ["Use the central daemon that maximizes the potential function";
       "for the next step (greedy). Performs |enabled| trials)"];

    mkopt args  ["--greedy-daemon";"-gd"]
      (Arg.Unit(fun () -> args.daemon <- DaemonType.Greedy))
      ["Use the daemon that maximizes the potential function";
       "for the next step (greedy). Performs 2^|enabled| trials) "];

    (*     mkopt args  ["--bad-daemon";"-bd"] ~arg:" <int>" *)
    (*       (Arg.Int (fun i -> args.daemon <- DaemonType.Bad i)) *)
    (*       ["Use a  daemon that tries  to maximize the  potential function, "; *)
    (*         "considering sub-graphs of a given maximal size"]; *)
    mkopt args  ["--init-search";"-is"]
      (Arg.Int(fun i -> args.init_search_max_trials <- Some i))
      ["Use local search algorithms to find an initial configuration that pessimize ";
       "the step number. The argument is the maximum number of trials to do the search. ";
       "Require the state_to_nums Algo.to_register field to be defined."]  ~arg:" <int>";

    mkopt args  ["--cores-nb";"-cn"]
      (Arg.Int(fun i -> args.cores_nb <- i))
      ["Number of cores to use during the simulation (default is 1)"];

   mkopt args ~hide:true ["--rif";"-rif"]
      (Arg.Unit(fun () -> args.rif <- true))
      ["Display only outputs on stdout (i.e., behave as a rif input file)"];

   mkopt args  ~hide:true  ["--no-data-file";"-nd"]
      (Arg.Unit(fun () -> args.no_data_file <- true))
      ["Do not generate any data file"];

    mkopt args  ["--seed";"-seed"]
      (Arg.Int(fun i -> Seed.set i))  ~arg:" <int>"
      ["Set the pseudo-random generator seed of build-in daemons (wins over --replay)"];
    
    mkopt args  ["--replay";"-replay"]
      (Arg.Unit(fun () -> Seed.replay_seed := true))
      ["Use the last generated seed to replay the last run"];

   mkopt args ~hide:true ["--gen-lutin-daemon";"-gld"]
      (Arg.Unit(fun () -> args.gen_lutin <- true))
      ["Generate Lutin daemons and exit"];

   mkopt args ~hide:true ["--gen-lustre-oracle-skeleton";"-glos"]
      (Arg.Unit(fun () -> args.gen_oracle <- true))
      ["Generate a Lustre oracle skeleton"];

   mkopt args ~hide:true ["--list-algos";"-algo"]
     (Arg.Unit(fun () -> args.output_algos <- true))
      ["Output the algo files used in the dot file and exit. "];
    
   mkopt args ~hide:false ["--gen-register";"-reg"]
     (Arg.Unit(fun () -> args.gen_register <- true))
      ["Generates the registering file and exit. "];
    
   mkopt args ~hide:true ["--dummy-input"]
      (Arg.Unit(fun () -> args.dummy_input <- true))
      ["Add a dummy input to sasa so that built-in daemon can be used from rdbg"];

    mkopt args ~hide:true ["--ignore-first-inputs"; "-ifi"]
      (Arg.Unit(fun () -> args.ifi <- true))
      ["[Deprecated] make sasa ignore its first input vector"];

    mkopt args ["--length";"-l"] ~arg:" <int>"
      (Arg.Int (fun i -> args.length <- i))
      ["Maximum number of steps to be done (" ^
       (string_of_int args.length) ^ " by default).\n"];

    mkopt args ["--version";"-version";"-v"]
      (Arg.Unit (fun _ ->
           (print_string (SasaVersion.str^"-"^SasaVersion.sha^"\n");
            flush stdout;exit 0)))
      ["Display the sasa version and exit."];

    mkopt args ~hide:true ["--ocaml-version"]
      (Arg.Unit (fun _ -> (print_string (Sys.ocaml_version^"\n"); flush stdout; exit 0)))
      ["Display the version ocaml version sasa was compiled with and exit."];

    mkopt args  ["--verbose";"-vl"] ~arg:" <int>"
      (Arg.Int (fun i -> args.verbose <- i))   ["Set the verbose level"];

    mkopt args ["--help";"-help"; "-h"]
      (Arg.Unit (fun _ -> print_usage (argv.(0)); options args stdout; exit 0))
      ["Display main options"];

    mkopt args ["--more";"-m"] (Arg.Unit (fun () -> more_options args stdout; exit 0)) 
      ["Display more options"]

  )

(* all unrecognized options are accumulated *)
let (add_other : t -> string -> unit) =
  fun opt s -> 
    opt._others <- s::opt._others

let current = ref 0;;
let first_line b = (
	try (
		let f = String.index b '\n' in
		String.sub b 0 f
	) with Not_found -> b
)
let file_notfound f = (
	prerr_string ("File not found: \""^f^"\"");
	prerr_newline ();
	myexit 1
)
let unexpected s = (
	prerr_string ("unexpected argument \""^s^"\"");
	prerr_newline ();
	myexit 1
)

let parse argv = (
  let save_current = !current in
  try (
    let args = make_args () in
    mkoptab argv args;
    Arg.parse_argv ~current:current argv args._args (add_other args) (usage_msg argv.(0));
    Functory.Cores.set_number_of_cores args.cores_nb;
    (List.iter 
       (fun f -> 
          if (String.sub f 0 1 = "-") then
            unexpected f 
          else if not (Sys.file_exists f) then
            file_notfound f
          else ()
       ) 
       args._others);
    current := save_current;
    args.topo <- (match args._others with
          [] -> 
          Printf.fprintf stderr "*** The topology file is missing in '%s'\n%s\n"
            (argv.(0)) (usage_msg argv.(0));  
          exit 2; 
        | x::_ -> x
      );
    args
  )
  with
  | Arg.Bad msg ->
    Printf.fprintf stderr " [sasa] Error when calling '%s': %s\n%s\n%!" (argv.(0))
      (first_line msg) (usage_msg argv.(0)); exit 2; 
  | Arg.Help msg -> 
    Printf.fprintf stdout "%s\n%s\n" msg (usage_msg argv.(0)); 
    exit 0
)
